Package: dir.expiry
Version: 1.18.0
Date: 2024-10-17
Title: Managing Expiration for Cache Directories
Description: Implements an expiration system for access to versioned directories.
    Directories that have not been accessed by a registered function within a certain time frame are deleted.
    This aims to reduce disk usage by eliminating obsolete caches generated by old versions of packages.
Authors@R: person("Aaron", "Lun", role=c("aut", "cre"), email="infinite.monkeys.with.keyboards@gmail.com")
License: GPL-3
Imports: utils, filelock
Suggests: rmarkdown, knitr, testthat, BiocStyle
biocViews: Software, Infrastructure
VignetteBuilder: knitr
RoxygenNote: 7.3.2
Encoding: UTF-8
git_url: https://git.bioconductor.org/packages/dir.expiry
git_branch: RELEASE_3_22
git_last_commit: 6d768b3
git_last_commit_date: 2025-10-29
Repository: Bioconductor 3.22
Date/Publication: 2025-10-29
NeedsCompilation: no
Packaged: 2025-10-30 03:37:50 UTC; biocbuild
Author: Aaron Lun [aut, cre]
Maintainer: Aaron Lun <infinite.monkeys.with.keyboards@gmail.com>

# Generated by roxygen2: do not edit by hand

export(clearDirectories)
export(lockDirectory)
export(touchDirectory)
export(unlockDirectory)
importFrom(filelock,lock)
importFrom(filelock,unlock)
importFrom(utils,packageVersion)

#' Touch a versioned directory
#'
#' Touch a versioned directory to indicate that it has been successfully accessed in the recent past.
#'
#' @param path String containing the path to a versioned directory.
#' The \code{dirname} should be the package cache while the \code{basename} should be a version number.
#' @param date A \link{Date} object containing the current date.
#' Only provided for testing.
#' @param force Logical scalar indicating whether to forcibly update the access date for \code{path}.
#'
#' @details
#' This function is used to update the last-accessed timestamp for a particular versioned directory.
#' The timestamp is stored in a \code{*_dir.expiry} stub file inside \code{path},
#' and is checked by \code{\link{clearDirectories}} to determine whether the directory is old enough for deletion.
#' We use an explicit timestamp instead of relying on POSIX access times as the latter may not be reliable indicators of genuine access
#' (e.g., during filesystem scans for viruses or to create backups) or may be turned off altogether for performance reasons.
# 
#' The \code{touchDirectory} function should be used in the following manner to ensure thread safety:
#' \enumerate{
#' \item The user should first call \code{\link{lockDirectory}(path)} before calling \code{touchDirectory(path)}.
#' This ensures that another process calling \code{\link{clearDirectories}} does not delete \code{path} while its access time is being updated.
#' \item The caller should then perform the desired operations on \code{path}.
#' This may be a read-only operation if the V lock is a shared lock or a read/write operation for an exclusive lock.
#' \item Once the operation has completed successfully, the caller should call \code{touchDirectory(path)}.
#' This ensures that the most recent timestamp is recorded, especially for long-running operations.
#' Multiple processes may safely call \code{touchDirectory(path)} when the V lock is shared.
#' \item Finally, the user should call \code{\link{unlockDirectory}(path)}.
#' This is typically wrapped in an \code{\link{on.exit}} to ensure that the locks are removed.
#' }
#'
#' For a given \code{path} and \code{version}, this function only modifies the files on its first call.
#' All subsequent calls with the same two arguments, in the same R session, and on the same day will have no effect.
#' This avoids unnecessary filesystem writes and locks when this function is repeatedly called.
#' Advanced users can force an update by setting \code{force=TRUE}.
#'
#' @return 
#' The \code{<version>_dir.expiry} stub file within \code{path} is updated/created with the current date.
#' A \code{NULL} is invisibly returned.
#'
#' @author Aaron Lun
#' @examples
#' # Creating the package cache.
#' cache.dir <- tempfile(pattern="expired_demo")
#' dir.create(cache.dir)
#'
#' # Creating the versioned subdirectory.
#' version <- package_version("1.11.0")
#' version.dir <- file.path(cache.dir, version)
#' lck <- lockDirectory(version.dir)
#' dir.create(version.dir)
#' 
#' # Setting the last access time.
#' touchDirectory(version.dir)
#' list.files(cache.dir)
#' readLines(file.path(cache.dir, "1.11.0_dir.expiry"))
#'
#' # Making sure we unlock it afterwards.
#' unlockDirectory(lck)
#' 
#' @seealso
#' \code{\link{lockDirectory}}, which should always be called before this function.
#'
#' @export
#' @importFrom utils packageVersion
#' @importFrom filelock lock unlock
touchDirectory <- function(path, date=Sys.Date(), force=FALSE) {
    out <- paste0(.unslash(path), expiry.suffix)

    # Skipping if we already touched it some time today. Note we have to 
    # check for the existence of the file just in case it got deleted by
    # clearDirectories (e.g., with limit=-Inf) in the meantime.
    if (.was_checked_today(path, touched.env) && !force && file.exists(out)) {
        return(invisible(NULL))
    }

    # Creating a lockfile for the actual expiry file, so that multiple
    # processes can touch it at the same time (e.g., if those processes were
    # otherwise read-only on the directory contents and thus held shared locks
    # via lockDirectory).
    lckfile <- paste0(out, lock.suffix)
    lckhandle <- lock(lckfile)
    on.exit(unlock(lckhandle))

    write.dcf(file=out, cbind(ExpiryVersion=as.character(packageVersion("dir.expiry")), AccessDate=as.integer(date)))

    invisible(NULL)
}

expiry.suffix <- "_dir.expiry"
touched.env <- new.env()
touched.env$status <- list()

.was_checked_today <- function(vpath, env) {
    info <- env$status[[vpath]]

    if (is.null(info)) {
        info <- list(last.checked=Sys.Date())
        env$status[[vpath]] <- info
        FALSE 
    } else {
        if (Sys.Date() == info$last.checked) {
            TRUE 
        } else {
            info$last.checked <- Sys.Date()
            env$status[[vpath]] <- info
            FALSE
        }
    }
}

.flush_cache <- function(env) {
    env$status <- list()
    invisible(NULL)
}

# This tests the makeCheckFunction() function.
# library(testthat); library(dir.expiry); source("test-checkfun.R")

test_that(".was_checked_today works as expected", {
    env <- new.env()
    env$status <- list()

    expect_false(dir.expiry:::.was_checked_today("rabbit", env))
    expect_true(dir.expiry:::.was_checked_today("rabbit", env))
    expect_true(dir.expiry:::.was_checked_today("rabbit", env))

    expect_false(dir.expiry:::.was_checked_today("dog", env))
    expect_true(dir.expiry:::.was_checked_today("dog", env))
    expect_true(dir.expiry:::.was_checked_today("dog", env))
})

test_that(".was_checked_today responds to the date", {
    env <- new.env()
    env$status <- list()
    expect_false(dir.expiry:::.was_checked_today("rabbit2", env))

    env$status$rabbit2$last.checked <- Sys.Date() -100
    expect_false(dir.expiry:::.was_checked_today("rabbit2", env))

    env$status$rabbit2$last.checked <- Sys.Date()
    expect_true(dir.expiry:::.was_checked_today("rabbit2", env))
    expect_true(dir.expiry:::.was_checked_today("rabbit2", env))
})

# This tests the touchDirectories function.
# library(testthat); library(dir.expiry); source("test-touch.R")

test_that("touchDirectories works as expected", {
    path <- tempfile(pattern="expired_demo")
    dir.create(path)
    version <- package_version("1.11.0")

    touchDirectory(file.path(path, version))
    expect_true(file.exists(file.path(path, "1.11.0_dir.expiry")))

    contents <- read.dcf(file.path(path, "1.11.0_dir.expiry"))
    expect_identical(as.character(as.integer(Sys.Date())), unname(contents[,"AccessDate"]))
})

test_that("touchDirectories skips work", {
    path <- tempfile(pattern="expired_demo")

    dir.create(path)
    version <- package_version("1.11.0")
    ver.dir <- file.path(path, version)

    touchDirectory(ver.dir)
    target <- file.path(path, "1.11.0_dir.expiry")
    expect_true(file.exists(target))

    # Doesn't bother to regenerate the file, because we skip this process entirely.
    write(file=target, "BLAH")
    touchDirectory(ver.dir)
    expect_identical(readLines(target), "BLAH")

    # Unless we force the issue.
    touchDirectory(ver.dir, force=TRUE)
    expect_match(readLines(target)[2], "AccessDate")

    # ... or if the file is deleted.
    unlink(target)
    touchDirectory(ver.dir)
    expect_true(file.exists(target))
})

# This tests the clearDirectories functionality.
# library(testthat); library(dir.expiry); source("test-clear.R")

test_that("clearDirectories works as expected", {
    path <- tempfile(pattern="expired_demo")
    dir.create(path)

    # Making an old directory.
    version <- package_version("1.11.0")
    ver.dir <- file.path(path, version)
    dir.create(ver.dir)
    touchDirectory(ver.dir, date=Sys.Date() - 100)

    # Making a more recent one without any clearing.
    version <- package_version("1.12.0")
    ver.dir <- file.path(path, version)
    dir.create(ver.dir)
    touchDirectory(ver.dir)

    earlier <- c("1.11.0", "1.11.0_dir.expiry")
    later <- c("1.12.0", "1.12.0_dir.expiry")
    expect_true(all(c(earlier, later) %in% list.files(path)))

    # Only deletes the expired directories.
    clearDirectories(path)

    expect_false(any(c(earlier, "1.11.0-00LOCK") %in% list.files(path)))
    expect_true(all(c(later, "1.12.0-00LOCK") %in% list.files(path))) 
    expect_true("central-00LOCK" %in% list.files(path))
})

test_that("clearDirectories responds to the environment variables", {
    path <- tempfile(pattern="expired_demo")
                                                                      
    dir.create(path)
    version <- package_version("1.11.0")
    ver.dir <- file.path(path, version)
    dir.create(ver.dir)
    touchDirectory(ver.dir, date=Sys.Date() - 100)

    expected <- c("1.11.0", "1.11.0_dir.expiry") 
    expect_true(all(expected %in% list.files(path)))

    # Doesn't delete anything if we pass in the right arguments.
    clearDirectories(path, limit=100)
    expect_true(all(expected %in% list.files(path)))

    old <- Sys.getenv("BIOC_DIR_EXPIRY_LIMIT", NA)
    Sys.setenv(BIOC_DIR_EXPIRY_LIMIT=100)
    clearDirectories(path)
    expect_true(all(expected %in% list.files(path)))

    if (is.na(old)) {
        Sys.unsetenv("BIOC_DIR_EXPIRY_LIMIT")
    } else {
        Sys.setenv(BIOC_DIR_EXPIRY_LIMIT=NA)
    }

    # Positive control. 
    clearDirectories(path, force=TRUE)
    expect_false(any(expected %in% list.files(path)))
})

test_that("clearDirectories doesn't delete the reference or newer versions", {
    path <- tempfile(pattern="expired_demo")
                                                                      
    dir.create(path)
    version <- package_version("1.11.0")
    ver.dir <- file.path(path, version)
    dir.create(ver.dir)
    touchDirectory(ver.dir, date=Sys.Date() - 100)

    version <- package_version("1.12.0")
    ver.dir <- file.path(path, version)
    dir.create(ver.dir)
    touchDirectory(ver.dir, date=Sys.Date() - 100)

    earlier <- c("1.11.0", "1.11.0_dir.expiry")
    later <- c("1.12.0", "1.12.0_dir.expiry")
    expected <- c(earlier, later)
    expect_true(all(expected %in% list.files(path)))

    # Doesn't delete the expired directories.
    clearDirectories(path, reference=package_version("1.11.0"))
    expect_true(all(expected %in% list.files(path)))

    # Keeps the reference. 
    clearDirectories(path, reference=package_version("1.12.0"), force=TRUE)
    expect_false(any(c(earlier, "1.11.0-00LOCK") %in% list.files(path)))
    expect_true(all(later %in% list.files(path))) # doesn't make the lock for the reference
})

#' Lock and unlock directories
#'
#' Lock and unlock the package and version directories for thread-safe processing.
#'
#' @inheritParams touchDirectory
#' @param ... For \code{lockDirectory}, further arguments to pass to \code{\link{lock}}.
#' 
#' For \code{unlockDirectory}, further arguments to pass to \code{\link{clearDirectories}}.
#' @param lock.info The list returned by \code{\link{lockDirectory}}.
#' @param clear Logical scalar indicating whether to remove expired versions via \code{\link{clearDirectories}}.
#' 
#' @return 
#' \code{lockDirectory} returns a list of locking information, including lock handles generated by the \pkg{filelock} package.
#'
#' \code{unlockDirectory} unlocks the handles generated by \code{lockDirectory}.
#' If \code{clear=TRUE}, versioned directories that have expired are removed by \code{\link{clearDirectories}}.
#' It returns a \code{NULL} invisibly.
#'
#' @details
#' \code{lockDirectory} actually creates two locks:
#' \itemize{
#' \item The first \dQuote{V} lock is applied to the versioned directory (i.e., \code{basename(path)}) within the package cache (i.e., \code{dirname(path)}).
#' This provides thread-safe read/write on its contents, protecting against other processes that want to write to the same versioned directory.
#' If the caller is only reading from \code{path}, they can set \code{exclusive=FALSE} in \code{...} to define a shared lock for concurrent reads across multiple processes.
#' \item The second \dQuote{P} lock is applied to the package cache and is always a shared lock, regardless of the contents of \code{...}.
#' This provides thread-safe access to the lock file used by the V lock, protecting it from deletion when the relevant directory expires in \code{\link{clearDirectories}}.
#' }
#' If \code{dirname(path)} does not exist, it will be created by \code{lockDirectory}.
#'
#' \code{\link{clearDirectories}} is called in \code{unlockDirectory} as the former needs to hold an exclusive lock on the package cache.
#' Thus, the clearing can only be performed after the P lock created by \code{lockDirectory} is released.
#'
#' @author Aaron Lun
#' 
#' @examples
#' # Creating the relevant directories.
#' cache.dir <- tempfile(pattern="expired_demo")
#' version <- package_version("1.11.0")
#' 
#' handle <- lockDirectory(file.path(cache.dir, version))
#' handle
#' unlockDirectory(handle)
#'
#' list.files(cache.dir)
#' 
#' @export
#' @importFrom filelock lock unlock
lockDirectory <- function(path, ...) {
    dir <- dirname(path)
    dir.create(dir, showWarnings=FALSE, recursive=TRUE)

    plock <- .plock_path(dir)
    second <- lock(plock, exclusive=FALSE) # define this first to protect the other lock.

    vlock <- .vlock_path(path)
    tryCatch({
        first <- lock(vlock, ...)
    }, error=function(e) {
        # Release the prior lock if acquiring this one fails.
        unlock(second)
        stop(e)
    })

    list(version=first, central=second, path=path)
}

lock.suffix <- "-00LOCK"

.unslash <- function(path) {
    sub("/$", "", path)
}

.vlock_path <- function(path) {
    paste0(.unslash(path), lock.suffix)
}

.plock_path <- function(dir) { 
    file.path(dir, paste0("central", lock.suffix))
}

#' @export
#' @importFrom filelock unlock
#' @rdname lockDirectory
unlockDirectory <- function(lock.info, clear=TRUE, ...) {
    # Unlock the first one while the second lock on the 'path' directory is
    # still in place to protect the lock file.
    unlock(lock.info$version)
    unlock(lock.info$central)

    if (clear) {
        path <- lock.info$path
        clearDirectories(dirname(path), reference=basename(path), ...)
    }
}

# Tests the lockDirectory function.
# library(testthat); library(dir.expiry); source("test-lock.R")

test_that("lockDirectory generates the requisite files", {
    base.path <- tempfile(pattern="expired_demo")
    dir.create(base.path)
    version <- package_version("1.11.0")

    handle <- lockDirectory(file.path(base.path, version))
    expect_identical(length(handle), 3L)

    expect_true(file.exists(file.path(base.path, paste0(version, "-00LOCK"))))
    expect_true(file.exists(file.path(base.path, "central-00LOCK")))
    unlockDirectory(handle)
})

test_that("lockDirectory works when the directory isn't yet generated", {
    base.path <- tempfile(pattern="expired_demo")
    version <- "1.0.0"
    handle <- lockDirectory(file.path(base.path, version))

    expect_true(file.exists(base.path))
    expect_true(file.exists(file.path(base.path, paste0(version, "-00LOCK"))))
    expect_true(file.exists(file.path(base.path, "central-00LOCK")))

    unlockDirectory(handle)
})

test_that("unlockDirectory calls the directory clearer", {
    path <- tempfile(pattern="expired_demo")

    dir.create(path)

    # Making one expired directory.
    version <- package_version("1.11.0")
    ver.dir <- file.path(path, version)
    dir.create(ver.dir)
    touchDirectory(ver.dir, date=Sys.Date() - 100)

    # Another versioned directory.
    version <- package_version("1.12.0")
    ver.dir <- file.path(path, version)
    dir.create(ver.dir)
    touchDirectory(ver.dir)

    # Locking and unlocking without calling the directory clearer. 
    lck <- lockDirectory(ver.dir)
    unlockDirectory(lck, clear=FALSE)

    all.files <- list.files(path)
    expect_true("1.12.0" %in% all.files)
    expect_true("1.12.0_dir.expiry" %in% all.files)
    expect_true("1.11.0" %in% all.files)
    expect_true("1.11.0_dir.expiry" %in% all.files)

    # And again, this time with the clearer.
    lck <- lockDirectory(ver.dir)
    unlockDirectory(lck)

    all.files <- list.files(path)
    expect_true("1.12.0" %in% all.files)
    expect_true("1.12.0_dir.expiry" %in% all.files)
    expect_false("1.11.0" %in% all.files)
    expect_false("1.11.0_dir.expiry" %in% all.files)
})

test_that("lockDirectory handles errors and sharedness correctly", {
    tmp <- file.path(tempfile(), "1.0.0")

    loc <- lockDirectory(tmp)
    expect_true(file.exists(paste0(tmp, "-00LOCK")))
    expect_error(lockDirectory(tmp, exclusive=FALSE), "exclusive")
    expect_null(unlockDirectory(loc))

    loc1 <- lockDirectory(tmp, exclusive=FALSE)
    loc2 <- lockDirectory(tmp, exclusive=FALSE)
    expect_error(lockDirectory(tmp, exclusive=TRUE), "shared")
    unlockDirectory(loc1)
    unlockDirectory(loc2)
})

#' Clear expired directories
#'
#' Remove versioned directories that have passed on expiration limit.
#'
#' @param dir String containing the path to a package cache containing any number of versioned directories.
#' @param reference A \link{package_version} specifying a reference version to be protected from deletion.
#' @param limit Integer scalar specifying the maximum number of days to have passed before a versioned directory expires.
#' @param force Logical scalar indicating whether to forcibly re-examine \code{dir} for expired versioned directories.
#' 
#' @return Expired directories are deleted and \code{NULL} is invisibly returned.
#'
#' @author Aaron Lun
#'
#' @details
#' This function checks the last access date in the \code{*_dir.expiry} files in \code{dir} (generated by \code{\link{touchDirectory}}).
#' If the last access date is too old, the corresponding subdirectory in \code{path} is treated as expired and is deleted.
#' The age threshold depends on \code{limit}, which defaults to the value of the environment variable \code{BIOC_DIR_EXPIRY_LIMIT}.
#' If this is not specified, it is set to 30 days.
#'
#' If \code{reference} is specified, any directory of that name is protected from deletion.
#' In addition, directories with version numbers greater than (or equal to) \code{reference} are not deleted,
#' even if their last access date was older than the specified \code{limit}.
#' This aims to favor the retention of newer versions, which is generally a sensible outcome when the aim is to stay up-to-date.
#' 
#' This function will acquire exclusive locks on the package cache directory and on each versioned directory before attempting to delete the latter.
#' Applications can achieve thread safety by calling \code{\link{lockDirectory}} prior to any operations on the versioned directory.
#' This ensures that \code{clearDirectories} will not delete a directory in use by another process, especially if the latter might update the last access time.
#'
#' By default, this function will remember the values of \code{dir} that were passed in previous calls,
#' and will avoid re-examining those same \code{dir}s for expired directories on the same day.
#' This avoids unnecessary file system queries and locks when this function is repeatedly called.
#' Advanced users can force a re-examination by setting \code{force=TRUE}.
#' 
#' @examples
#' # Creating the package cache.
#' cache.dir <- tempfile(pattern="expired_demo")
#'
#' # Creating an older versioned directory.
#' version <- package_version("1.11.0")
#' version.dir <- file.path(cache.dir, version)
#'
#' lck <- lockDirectory(version.dir)
#' dir.create(version.dir)
#' touchDirectory(version.dir, date=Sys.Date() - 100)
#' unlockDirectory(lck, clear=FALSE) # manually clear below.
#'
#' list.files(cache.dir)
#'
#' # Clearing them out.
#' clearDirectories(cache.dir)
#' list.files(cache.dir)
#' 
#' @seealso
#' \code{\link{touchDirectory}}, which calls this function automatically when \code{clear=TRUE}.
#' @export
clearDirectories <- function(dir, reference=NULL, limit=NULL, force=FALSE) {
    if (.was_checked_today(dir, cleared.env) && !force) {
        return(invisible(NULL))
    }

    if (is.null(limit)) {
        limit <- Sys.getenv("BIOC_DIR_EXPIRY_LIMIT", "30")
        limit <- as.integer(limit)
    } 

    # Unlike lockDirectory, this is exclusive as we will be deleting the lock
    # files; no point allowing other processes to touch them in the meantime.
    # We put the locking here so as to ensure that the files don't disappear
    # between the list.files() and the actual removal.
    plock <- .plock_path(dir)
    p <- lock(plock) 
    on.exit(unlock(p))

    pattern <- paste0(expiry.suffix, "$")
    all.files <- list.files(dir, pattern=pattern)
    if (!is.null(reference)) {
        all.files <- setdiff(all.files, paste0(as.character(reference), expiry.suffix))
        if (is.character(reference)) {
            reference <- package_version(reference)
        }
    }

    current <- as.integer(Sys.Date())
    for (x in all.files) {
        version <- sub(pattern, "", x)
        .delete_versioned_directory(dir, version=version, expfile=x, date=current, reference=reference, limit=limit)
    }

    invisible(NULL)
}

cleared.env <- new.env()
cleared.env$status <- list()

#' @importFrom filelock lock unlock
.delete_versioned_directory <- function(dir, version, expfile, date, reference, limit) {
    # Protect against simultaneous accesses to the targeted directory,
    # assuming users called lockDirectory() before touchDirectory().
    path <- file.path(dir, version)
    vlock <- .vlock_path(path)
    V <- lock(vlock)

    deleted <- FALSE
    on.exit({ 
        unlock(V)
        if (deleted) {
            unlink(vlock)
        }
    }, add=TRUE, after=FALSE)

    acc.path <- file.path(dir, expfile)
    last.used <- as.integer(read.dcf(acc.path)[,"AccessDate"])
    diff <- date - last.used

    if (diff > limit && (is.null(reference) || reference > package_version(version))) {
        unlink(acc.path, force=TRUE)
        unlink(paste0(acc.path, lock.suffix), force=TRUE)
        unlink(path, recursive=TRUE, force=TRUE)
        deleted <- TRUE
    }
}

## ----echo=FALSE---------------------------------------------------------------
knitr::opts_chunk$set(error=FALSE, message=FALSE, warnings=FALSE)
library(BiocStyle)

## -----------------------------------------------------------------------------
cache.path <- tempfile(pattern="expired_demo")
dir.create(cache.path)

version <- package_version("1.0.0")
version.dir <- file.path(cache.path, version)
dir.create(version.dir)

## -----------------------------------------------------------------------------
version.dir

## -----------------------------------------------------------------------------
library(dir.expiry)
touchDirectory(version.dir)

## -----------------------------------------------------------------------------
list.files(cache.path)
cat(readLines(file.path(cache.path, "1.0.0_dir.expiry")), sep="\n")

## -----------------------------------------------------------------------------
v <- package_version("1.0.0")
v.dir <- file.path(cache.path, v)

lock <- lockDirectory(v.dir)
# on.exit(unlockDirectory(lock)), in a function context.

# Do stuff with the versioned cache 'v.dir' here...
dir.create(v.dir, showWarnings=FALSE)

# Finally, touch the directory on successful completion.
touchDirectory(v.dir)

## -----------------------------------------------------------------------------
unlockDirectory(lock) 

## -----------------------------------------------------------------------------
cache.path <- tempfile(pattern="expired_demo")

old.version <- package_version("0.99.0")
old.version.dir <- file.path(cache.path, old.version)

lck <- lockDirectory(old.version.dir)
dir.create(old.version.dir)
touchDirectory(old.version.dir, date=Sys.Date() - 100)
unlockDirectory(lck, clear=FALSE)

list.files(cache.path)

## -----------------------------------------------------------------------------
new.version <- package_version("1.0.0")
new.version.dir <- file.path(cache.path, new.version)

lck <- lockDirectory(new.version.dir)
dir.create(new.version.dir)
touchDirectory(new.version.dir)
unlockDirectory(lck)

list.files(cache.path)

## -----------------------------------------------------------------------------
new.version2 <- package_version("1.0.1")
new.version.dir2 <- file.path(cache.path, new.version2)

# Newer version but earlier access.
lck2 <- lockDirectory(new.version.dir2)
dir.create(new.version.dir2)
touchDirectory(new.version.dir2, date=Sys.Date() - 100) 
unlockDirectory(lck2)

# Re-accessing the older version.
lck <- lockDirectory(new.version.dir)
touchDirectory(new.version.dir)
unlockDirectory(lck)

list.files(cache.path)

## -----------------------------------------------------------------------------
sessionInfo()

---
title: Managing expiration of versioned subdirectories
author: 
- name: Aaron Lun
  email: infinite.monkeys.with.keyboards@gmail.com
date: "Revised: September 4, 2022"
output:
  BiocStyle::html_document:
    toc_float: true
package: dir.expiry
vignette: >
  %\VignetteIndexEntry{Managing directory expiration}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}    
---

```{r, echo=FALSE}
knitr::opts_chunk$set(error=FALSE, message=FALSE, warnings=FALSE)
library(BiocStyle)
```

# Background

Several Bioconductor packages (e.g., `r Biocpkg("basilisk")`, `r Biocpkg("rebook")`) use external cache directories to manage its resources.
This typically involves creating a package cache directory in which versioned directories are created:

```
~/.cache/
  basilisk/  # package cache directory
    0.99.0/  # versioned directory
    1.0.1/
    1.1.3/
```

The use of separate versioned caches allows the simultaneous use of multiple R installations with different versions of the same package.
However, it requires some housekeeping to remove stale caches and avoid an large increase in disk usage - especially for packages with frequent version updates.

One might think of using POSIX access time but this is not a reliable indicator of whether a cache is truly being used by the package.
For example, external processes like virus scans, `find -exec grep` and so on can update the last access time without actually using the cache's contents.
Some filesystem administrators may even turn off access time tracking for performance.

# Setting the access time

`r Biocpkg("dir.expiry")` implements its own tracker to monitor the last access time for a particular cache version.
To illustrate, let's set up a versioned cache directory.

```{r}
cache.path <- tempfile(pattern="expired_demo")
dir.create(cache.path)

version <- package_version("1.0.0")
version.dir <- file.path(cache.path, version)
dir.create(version.dir)
```

The path to this versioned directory should contain the version number as the `basename` and the package cache directory as the `dirname`.
The expectation is that there may be multiple versioned directories nested within a single package's cache directory.

```{r}
version.dir
```

Assume that we successfully accessed the versioned directory (for some arbitrary definition of success).
We can then update the last access time by calling the `touchDirectory()` function with the path to the versioned directory:

```{r}
library(dir.expiry)
touchDirectory(version.dir)
```

Doing so creates or updates an `*_dir.expiry` file containing the last access date as an integer.
(This implies that the process creating the cache should not make any of its own files ending with `*_dir.expiry-info`, lest they be confused with `r Biocpkg("dir.expiry")`'s files.)
Only directories with Bioconductor-style versions should be present in the package cache directory.

```{r}
list.files(cache.path)
cat(readLines(file.path(cache.path, "1.0.0_dir.expiry")), sep="\n")
```

# Maintaining thread safety

`r Biocpkg("dir.expiry")` makes extensive use of file locks (via the `r CRANpkg("filelock")` package) to avoid deleting caches that might be in use by other R processes.
Developers are strongly advised to lock their directories of interest before doing anything with them, including any calls to `touchDirectory()`.
This should be done using the `lockDirectory()` function:

```{r}
v <- package_version("1.0.0")
v.dir <- file.path(cache.path, v)

lock <- lockDirectory(v.dir)
# on.exit(unlockDirectory(lock)), in a function context.

# Do stuff with the versioned cache 'v.dir' here...
dir.create(v.dir, showWarnings=FALSE)

# Finally, touch the directory on successful completion.
touchDirectory(v.dir)
```

In real functions, `on.exit()` is the preferred approach for releasing the locks created by `lockDirectory()`.
In this vignette, though, we will just call this function manually.

```{r}
unlockDirectory(lock) 
```

For read-only applications, developers may wish to set `exclusive=FALSE` in `lockDirectory()`.
This allows multiple processes to read from the versioned cache at the same time, only being blocked when one process needs to perform a write operation.

# Deleting old caches

`unlockDirectory()` will automatically delete all expired versions that it finds in the package cache directory.
By default, this is defined as all directories that were last touched more than 30 days ago.
To illustrate, let's create another versioned directories and pretend it was created 100 days ago:

```{r}
cache.path <- tempfile(pattern="expired_demo")

old.version <- package_version("0.99.0")
old.version.dir <- file.path(cache.path, old.version)

lck <- lockDirectory(old.version.dir)
dir.create(old.version.dir)
touchDirectory(old.version.dir, date=Sys.Date() - 100)
unlockDirectory(lck, clear=FALSE)

list.files(cache.path)
```

Locking and unlocking a more recent versioned directory will delete the files associated with the older version.
This includes the versioned directory itself as well as any lock and expiry files.

```{r}
new.version <- package_version("1.0.0")
new.version.dir <- file.path(cache.path, new.version)

lck <- lockDirectory(new.version.dir)
dir.create(new.version.dir)
touchDirectory(new.version.dir)
unlockDirectory(lck)

list.files(cache.path)
```

However, the converse is not true; locking/unlocking the older version will never delete files associated with a newer version.
This favors retention of caches corresponding to later versions, which is generally a reasonable outcome. 

```{r}
new.version2 <- package_version("1.0.1")
new.version.dir2 <- file.path(cache.path, new.version2)

# Newer version but earlier access.
lck2 <- lockDirectory(new.version.dir2)
dir.create(new.version.dir2)
touchDirectory(new.version.dir2, date=Sys.Date() - 100) 
unlockDirectory(lck2)

# Re-accessing the older version.
lck <- lockDirectory(new.version.dir)
touchDirectory(new.version.dir)
unlockDirectory(lck)

list.files(cache.path)
```

Under the hood, `unlockDirectory()` calls `clearDirectories()` to destroy the expired versioned directories.
Users can tune the expiration threshold by setting the number of days in the `limit=` argument or the `BIOC_DIR_EXPIRY_LIMIT` environment variable.

When deleting old versions, `clearDirectories()` will attempt to acquire an exclusive lock on the corresponding directories.
As long as other processes call `lockDirectory()` before working on them,
we ensure that `clearDirectories()` does not inadvertently delete a versioned directory that another process is simultaneously operating on.

# Session information {-}

```{r}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/touchDirectory.R
\name{touchDirectory}
\alias{touchDirectory}
\title{Touch a versioned directory}
\usage{
touchDirectory(path, date = Sys.Date(), force = FALSE)
}
\arguments{
\item{path}{String containing the path to a versioned directory.
The \code{dirname} should be the package cache while the \code{basename} should be a version number.}

\item{date}{A \link{Date} object containing the current date.
Only provided for testing.}

\item{force}{Logical scalar indicating whether to forcibly update the access date for \code{path}.}
}
\value{
The \code{<version>_dir.expiry} stub file within \code{path} is updated/created with the current date.
A \code{NULL} is invisibly returned.
}
\description{
Touch a versioned directory to indicate that it has been successfully accessed in the recent past.
}
\details{
This function is used to update the last-accessed timestamp for a particular versioned directory.
The timestamp is stored in a \code{*_dir.expiry} stub file inside \code{path},
and is checked by \code{\link{clearDirectories}} to determine whether the directory is old enough for deletion.
We use an explicit timestamp instead of relying on POSIX access times as the latter may not be reliable indicators of genuine access
(e.g., during filesystem scans for viruses or to create backups) or may be turned off altogether for performance reasons.
The \code{touchDirectory} function should be used in the following manner to ensure thread safety:
\enumerate{
\item The user should first call \code{\link{lockDirectory}(path)} before calling \code{touchDirectory(path)}.
This ensures that another process calling \code{\link{clearDirectories}} does not delete \code{path} while its access time is being updated.
\item The caller should then perform the desired operations on \code{path}.
This may be a read-only operation if the V lock is a shared lock or a read/write operation for an exclusive lock.
\item Once the operation has completed successfully, the caller should call \code{touchDirectory(path)}.
This ensures that the most recent timestamp is recorded, especially for long-running operations.
Multiple processes may safely call \code{touchDirectory(path)} when the V lock is shared.
\item Finally, the user should call \code{\link{unlockDirectory}(path)}.
This is typically wrapped in an \code{\link{on.exit}} to ensure that the locks are removed.
}

For a given \code{path} and \code{version}, this function only modifies the files on its first call.
All subsequent calls with the same two arguments, in the same R session, and on the same day will have no effect.
This avoids unnecessary filesystem writes and locks when this function is repeatedly called.
Advanced users can force an update by setting \code{force=TRUE}.
}
\examples{
# Creating the package cache.
cache.dir <- tempfile(pattern="expired_demo")
dir.create(cache.dir)

# Creating the versioned subdirectory.
version <- package_version("1.11.0")
version.dir <- file.path(cache.dir, version)
lck <- lockDirectory(version.dir)
dir.create(version.dir)

# Setting the last access time.
touchDirectory(version.dir)
list.files(cache.dir)
readLines(file.path(cache.dir, "1.11.0_dir.expiry"))

# Making sure we unlock it afterwards.
unlockDirectory(lck)

}
\seealso{
\code{\link{lockDirectory}}, which should always be called before this function.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lockDirectory.R
\name{lockDirectory}
\alias{lockDirectory}
\alias{unlockDirectory}
\title{Lock and unlock directories}
\usage{
lockDirectory(path, ...)

unlockDirectory(lock.info, clear = TRUE, ...)
}
\arguments{
\item{path}{String containing the path to a versioned directory.
The \code{dirname} should be the package cache while the \code{basename} should be a version number.}

\item{...}{For \code{lockDirectory}, further arguments to pass to \code{\link{lock}}.

For \code{unlockDirectory}, further arguments to pass to \code{\link{clearDirectories}}.}

\item{lock.info}{The list returned by \code{\link{lockDirectory}}.}

\item{clear}{Logical scalar indicating whether to remove expired versions via \code{\link{clearDirectories}}.}
}
\value{
\code{lockDirectory} returns a list of locking information, including lock handles generated by the \pkg{filelock} package.

\code{unlockDirectory} unlocks the handles generated by \code{lockDirectory}.
If \code{clear=TRUE}, versioned directories that have expired are removed by \code{\link{clearDirectories}}.
It returns a \code{NULL} invisibly.
}
\description{
Lock and unlock the package and version directories for thread-safe processing.
}
\details{
\code{lockDirectory} actually creates two locks:
\itemize{
\item The first \dQuote{V} lock is applied to the versioned directory (i.e., \code{basename(path)}) within the package cache (i.e., \code{dirname(path)}).
This provides thread-safe read/write on its contents, protecting against other processes that want to write to the same versioned directory.
If the caller is only reading from \code{path}, they can set \code{exclusive=FALSE} in \code{...} to define a shared lock for concurrent reads across multiple processes.
\item The second \dQuote{P} lock is applied to the package cache and is always a shared lock, regardless of the contents of \code{...}.
This provides thread-safe access to the lock file used by the V lock, protecting it from deletion when the relevant directory expires in \code{\link{clearDirectories}}.
}
If \code{dirname(path)} does not exist, it will be created by \code{lockDirectory}.

\code{\link{clearDirectories}} is called in \code{unlockDirectory} as the former needs to hold an exclusive lock on the package cache.
Thus, the clearing can only be performed after the P lock created by \code{lockDirectory} is released.
}
\examples{
# Creating the relevant directories.
cache.dir <- tempfile(pattern="expired_demo")
version <- package_version("1.11.0")

handle <- lockDirectory(file.path(cache.dir, version))
handle
unlockDirectory(handle)

list.files(cache.dir)

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clearDirectories.R
\name{clearDirectories}
\alias{clearDirectories}
\title{Clear expired directories}
\usage{
clearDirectories(dir, reference = NULL, limit = NULL, force = FALSE)
}
\arguments{
\item{dir}{String containing the path to a package cache containing any number of versioned directories.}

\item{reference}{A \link{package_version} specifying a reference version to be protected from deletion.}

\item{limit}{Integer scalar specifying the maximum number of days to have passed before a versioned directory expires.}

\item{force}{Logical scalar indicating whether to forcibly re-examine \code{dir} for expired versioned directories.}
}
\value{
Expired directories are deleted and \code{NULL} is invisibly returned.
}
\description{
Remove versioned directories that have passed on expiration limit.
}
\details{
This function checks the last access date in the \code{*_dir.expiry} files in \code{dir} (generated by \code{\link{touchDirectory}}).
If the last access date is too old, the corresponding subdirectory in \code{path} is treated as expired and is deleted.
The age threshold depends on \code{limit}, which defaults to the value of the environment variable \code{BIOC_DIR_EXPIRY_LIMIT}.
If this is not specified, it is set to 30 days.

If \code{reference} is specified, any directory of that name is protected from deletion.
In addition, directories with version numbers greater than (or equal to) \code{reference} are not deleted,
even if their last access date was older than the specified \code{limit}.
This aims to favor the retention of newer versions, which is generally a sensible outcome when the aim is to stay up-to-date.

This function will acquire exclusive locks on the package cache directory and on each versioned directory before attempting to delete the latter.
Applications can achieve thread safety by calling \code{\link{lockDirectory}} prior to any operations on the versioned directory.
This ensures that \code{clearDirectories} will not delete a directory in use by another process, especially if the latter might update the last access time.

By default, this function will remember the values of \code{dir} that were passed in previous calls,
and will avoid re-examining those same \code{dir}s for expired directories on the same day.
This avoids unnecessary file system queries and locks when this function is repeatedly called.
Advanced users can force a re-examination by setting \code{force=TRUE}.
}
\examples{
# Creating the package cache.
cache.dir <- tempfile(pattern="expired_demo")

# Creating an older versioned directory.
version <- package_version("1.11.0")
version.dir <- file.path(cache.dir, version)

lck <- lockDirectory(version.dir)
dir.create(version.dir)
touchDirectory(version.dir, date=Sys.Date() - 100)
unlockDirectory(lck, clear=FALSE) # manually clear below.

list.files(cache.dir)

# Clearing them out.
clearDirectories(cache.dir)
list.files(cache.dir)

}
\seealso{
\code{\link{touchDirectory}}, which calls this function automatically when \code{clear=TRUE}.
}
\author{
Aaron Lun
}
